{-# LANGUAGE DeriveFunctor #-}
{-# LANGUAGE DeriveFoldable #-}
{-# LANGUAGE DeriveTraversable #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE OverloadedStrings #-}
module ArbitraryEnzymeTree
  ( EnzymeTree
  , getTreeSize
  , getElement
  , getShape
  , getSubTrees
  , treeFromList
  , generateTreeFromList
  , generateDotFromTree
  ) where

import Test.QuickCheck
import GHC.Generics (Generic())
import Text.Pretty.Simple
import Data.List
import Data.String (IsString(..))
import Data.Monoid

data TreeShape = Root
               { treeSize :: Int
               , treeSubSizes :: [TreeShape]
               } deriving (Show, Eq)

data EnzymeTree s a = EnzymeTree TreeShape a [EnzymeTree TreeShape a] deriving (Show, Eq, Generic, Functor, Foldable, Traversable)

instance Applicative (EnzymeTree s) where
    pure a = EnzymeTree (Root 1 []) a []
    EnzymeTree s f fs <*> EnzymeTree _ a as = EnzymeTree s (f a) (zipWith (<*>) fs as)

instance Arbitrary a => Arbitrary (EnzymeTree s a) where
    arbitrary = sized arbitrarySizedEnzymeTree

arbitrarySizedEnzymeTree :: Arbitrary a => Int -> Gen (EnzymeTree s a)
arbitrarySizedEnzymeTree m = do
  t <- arbitrary
  n <- choose (m `div` 4, m)
  ts <- if n == 0 then return []
                  else sequenceA $ replicate (2*n) (arbitrarySizedEnzymeTree (n `div` 2))
  let (_,ts') = foldr (go m) (0,[]) ts
      go :: Int -> EnzymeTree s a -> (Int, [EnzymeTree s a]) -> (Int, [EnzymeTree s a])
      go m x (s,ts)
        | m == 0             = (s,ts)
        | s + getTreeSize x > m = (s,ts)
        | otherwise          = (s + getTreeSize x, x:ts)

      sz = sum $ getTreeSize <$> ts'
      ss = (\(EnzymeTree a _ _) -> a) <$> ts'
  return (EnzymeTree (Root (sz+1) ss) t ts')

getTreeSize :: EnzymeTree s a -> Int
getTreeSize (EnzymeTree (Root a _) _ _) = a

getShape :: EnzymeTree s a -> TreeShape
getShape (EnzymeTree s _ _) = s

getElement :: EnzymeTree s a -> a
getElement (EnzymeTree _ e _) = e

getSubTrees :: EnzymeTree s a -> [EnzymeTree TreeShape a]
getSubTrees (EnzymeTree _ _ s) = s

treeFromList :: TreeShape -> [a] -> EnzymeTree s a
treeFromList (Root n ns) (a:as) = EnzymeTree (Root n ns) a (unfoldr go (ns,as))
  where
      go :: ([TreeShape],[a]) -> Maybe (EnzymeTree s a,([TreeShape],[a]))
      go (n:ns,as) = Just (treeFromList n (take (treeSize n) as), (ns, drop (treeSize n) as))
      go ([],_)    = Nothing

generateTreeFromList :: Int -> [a] -> IO (EnzymeTree s a)
generateTreeFromList s a = flip treeFromList a . getShape <$> generate (resize s arbitrary :: Gen (EnzymeTree s ()))


generateDotFromTree :: (Show a, IsString b, Monoid b) => b -> EnzymeTree s a -> b
generateDotFromTree name t = "digraph " <> name <> " {\n"
                             <> generateDotFromTree' t
                             <> "}\n"
  where
    generateDotFromTree' :: (Show a, IsString b, Monoid b) => EnzymeTree s a -> b
    generateDotFromTree' (EnzymeTree _ num ns) = 
        "    " <> wrap (ts num) <> " -> { "
                   <> mconcat (intersperse " "  (wrap . ts . getElement <$> ns))
               <> " };\n"
               <> mconcat (generateDotFromTree' <$> ns)
      where 
        ts = fromString . show
        wrap x = "\"" <> x <> "\""
